/* Main source file.

   OperStats - Copyright (c) 2000-2001 Darcy Grexton
   Contact: cygnus@uniserve.ca, skold @ HabberNet, ICQ #6285178

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program (see the file COPYING); if not, write to the
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "../inc/operstats.h"
#include "../inc/version.h"

int servsock = -1;
int servcnt = 0, usercnt = 0, chancnt = 0;
int servstat = 0, userstat = 0, chanstat = 0, speedstat;
int dailyservstat = 0, dailyuserstat = 0, dailychanstat = 0;
int maxservcnt = 0, maxusercnt = 0, maxchancnt = 0;
int debuglevel = 0, donereset = 0;
time_t maxservtime, maxusertime, maxchantime, last_sample, last_update;
time_t last_check, dailyservtime, dailyusertime, dailychantime;
char inbuf[BUFSIZE], *quitmsg = NULL;
long runflags = 0;
static int processing = 0;
static jmp_buf panic_jmp;

/* If we get a weird signal, come here. */
static void sighandler (int signum)
{
    if (runflags & RUN_STARTED)
    {
	/* Rehash */
	if (signum == SIGHUP)
	{
	    signal (SIGHUP, SIG_IGN);

            log ("Got SIGHUP, reloading %s file", CONFIGFILE);

	    if (operserv_on == TRUE)
		do_rehash (s_OperServ, "system console");
	    else
		do_rehash (s_StatServ, "system console");
        }
	else if (signum == SIGTERM)
	{
	    runflags |= (RUN_SIGTERM);
	    signal (SIGTERM, SIG_IGN);
	    return;
	}
	else if (processing == 1)
	{
	    log ("PANIC! Buffer = %s", inbuf);
	    globops (NULL, "Fatal error. See logfile. [sig%d]", signum);
	    quitmsg = "Fatal error. See logfile.";
	}
    }
    if (signum == SIGUSR1 || !(quitmsg = malloc (BUFSIZE)))
    {
	quitmsg = "Out of memory!";
	runflags |= RUN_QUITTING;
    }
    else
    {
	snoop (s_OperServ, "Fatal Error: OperStats terminating on signal %d",
	    signum);
	snprintf (quitmsg, BUFSIZE, "OperStats terminating on signal %d.",
	    signum);
	runflags |= RUN_QUITTING;
    }
    if (runflags & RUN_STARTED)
	longjmp (panic_jmp, 1);
    else
    {
	log (ERR_QUIT, quitmsg);
	exit (1);
    }
}

/* Command-line help message */
void do_help (char *prog)
{
printf("
OperStats - Copyright (c) 2000-2001 Darcy Grexton

Syntax: %s [parms] [options]

-help                This message
-debug [0|1|2|3]     Enable debug mode - log to debug.log
                         Level 0: No debug logging (Off)
                         Level 1: Server<->server traffic logged
                         Level 2: (Some) Function calls logged
                         Level 3: Server<->server traffic and (some) function
				  calls logged
-live                Don't fork (log screen + log file).

", prog);
}

/* Count our servers and clients so the counts we show are accurate. */
int servcount (void)
{
    int tmp = 0;

    /* We have seperate counts for statistics purposes since we need
       the counts we use internally to be accurate, not including us.
     */
    if (operserv_on == TRUE)
	++tmp;
    if (statserv_on == TRUE)
        ++tmp;
    if (globalnoticer_on == TRUE)
	++tmp;

    return tmp;
}

/* Main routine .. :P */
int main (int ac, char **av, char **envp)
{
    FILE *pid_file;
    char *s;
    int i;

    if (chdir (OPERSTATS_DIR) < 0)
    {
	perror (OPERSTATS_DIR);
	return 20;
    }

    /* Parse our command line options */
    for (i = 1; i < ac; ++i)
    {
	s = av[i];
	if (*s == '-' || *s == '/')
	{
	    ++s;
	    if (!stricmp (s, "-help") || !stricmp (s, "?") ||
		  !stricmp (s, "help"))
	    {
		do_help (av[0]);
		return 0;
	    }
	    else if (!stricmp (s, "debug"))
	    {
		if (++i >= ac)
		{
		    fprintf (stderr, "Error: -debug requires a parameter.\n");
		    return -1;
		}
		s = av[i];
		debuglevel = atoi (s);
		open_debug ();
		debug ("-----------------------------------------------------");
	    }
	    else if (!stricmp (s, "live") || !stricmp (s, "nofork"))
		runflags |= RUN_LIVE;
	}
    }

    me.since = maxservtime = maxusertime = maxchantime = time (NULL);
        
    while (!(runflags & RUN_TERMINATING))
    {
	runflags |= RUN_STARTING;
	me.reset = time (NULL);
	last_sample = time (NULL);
	last_check = time (NULL);

	if (quitmsg)
	    free (quitmsg);

	quitmsg = NULL;
	servsock = -1;

	/* Open the logfile */
	open_log ();

	/* Fork into the background */
	if (!(runflags & RUN_LIVE))
	{
	    if ((i = fork ()) < 0)
	    {
		log ("Cannot fork!");
		return -1;
	    }
	    else if (i != 0)
		exit (0);

	    if (setpgid (0, 0) < 0)
	    {
		log ("setpgid()");
		return -1;
	    }
	}

	pid_file = fopen (PIDFILE, "w");

	if (pid_file)
	{
	    fprintf (pid_file, "%d\n", getpid ());
	    fclose (pid_file);
	}
	else
	    log (ERR_CANT_WRITE_PID);

	/* Write to the log so we know we're on a new session */
	log ("-----------------------------------------------------");
	log ("OperStats v%s boot complete.", version);

	/* Set signal handlers */
	signal (SIGINT, sighandler);
	signal (SIGTERM, sighandler);
	signal (SIGPIPE, SIG_IGN);
	signal (SIGQUIT, sighandler);
#ifndef DUMP_CORE
	signal (SIGSEGV, sighandler);
#endif
	signal (SIGBUS, sighandler);
	signal (SIGHUP, sighandler);
	signal (SIGILL, sighandler);
	signal (SIGTRAP, sighandler);
	signal (SIGIOT, sighandler);
	signal (SIGFPE, sighandler);
	signal (SIGALRM, SIG_IGN);
	signal (SIGUSR2, SIG_IGN);
	signal (SIGCHLD, SIG_IGN);
	signal (SIGWINCH, SIG_IGN);
	signal (SIGTTIN, SIG_IGN);
	signal (SIGTTOU, SIG_IGN);
	signal (SIGTSTP, SIG_IGN);
	signal (SIGUSR1, sighandler);

	/* Load our conf */
	load_conf ();

	/* Sanity checks */
	check_conf ();

	/* Set up OperServ stuff */
	if (operserv_on == TRUE)
	    load_os_dbase ();

	/* Set up StatServ stuff */
	if (statserv_on == TRUE)
	{
	    load_ss_dbase ();
	    load_tlds ();
	}

	/* Count our stuff */
	userstat += servcount ();
	servstat++;
	maxusercnt = userstat;

	log (RPL_CONNECTING, me.uplink, me.port);
	servsock = conn (me.uplink, me.port, me.bindhost, me.bindport);

	if (servsock < 0)
	{
	    log (ERR_CANT_CONNECT, me.uplink, me.port);
	    disconn (servsock);
	    close_log();
	    if (debuglevel)
		close_debug();
	    remove (PIDFILE);
	    return 0;
	}

	log (RPL_CONNECTED, me.uplink);
	log (RPL_LOGGING_IN);

	/* Send our initial login info to the server. */

#ifdef UNREAL3
	send_cmd (NULL, "PROTOCTL NOQUIT TOKEN VL");
#elif defined (BAHAMUT)
	send_cmd (NULL, "PROTOCTL NOQUIT");
#else
	send_cmd (NULL, "PROTOCTL NOQUIT TOKEN");
#endif

#ifdef BAHAMUT
	send_cmd (NULL, "PASS %s :TS", me.pass);
#else
	send_cmd (NULL, "PASS :%s", me.pass);
#endif

#ifdef UNREAL3
	send_cmd (NULL, "SERVER %s 1 :U0-*-%d %s", me.name, unreal_numeric,
	    me.desc);
#else
	send_cmd (NULL, "SERVER %s 1 :%s", me.name, me.desc);
#endif

#ifdef BAHAMUT
	send_cmd (NULL, "SVINFO 3 3 0 :%lu", time (NULL));
	send_cmd (NULL, "CAPAB: TS3");
#endif

	sgets2 (inbuf, sizeof (inbuf), servsock);

	if (!strncmp (inbuf, "ERROR", 5))
	{
	    quitmsg = inbuf;
	    log (ERR_ERROR, inbuf);
	    disconn (servsock);
	    servsock = -1;
	    close_log();
	    if (debuglevel)
		close_debug();
	    remove (PIDFILE);
	}

	log (RPL_LOGGED_IN);

	/* Now bring on our clients.. */
	if (operserv_on == TRUE)
	{
	    intro_user (s_OperServ, os.user, os.host, os.real, os.mode);
	    os.isonline = 1;
	}

	if (statserv_on == TRUE)
	{
	    intro_user (s_StatServ, ss.user, ss.host, ss.real, ss.mode);
	    ss.isonline = 1;
	}

	if (globalnoticer_on == TRUE)
	{
	    intro_user (s_GlobalNoticer, gn.user, gn.host, gn.real, gn.mode);
	    gn.isonline = 1;
	}

	last_update = time (NULL);
	last_check = time (NULL);

	/* Process the initial buffer.. */
	processing = 1;
	process ();
	processing = 0;

	setjmp (panic_jmp);
	runflags &= ~RUN_STARTING;
	runflags |= RUN_STARTED;

	/* Main loop. */
	while (!(runflags & RUN_QUITTING))
	{
	    time_t curtime = time (NULL);

	    /* Reset the stat every second */
	    if (curtime > last_sample)
	    {
		if (speedsample > speedstat)
		    speedstat = speedsample;
		speedsample = 0;
		last_sample = curtime;
	    }

	    /* Check various timeouts */
	    if (curtime - last_check > 5)
	    {
		/* Check various StatServ stuff */
		if (statserv_on == TRUE)
		{
		    struct tm tm;
		    static char timebuf[256];

		    checkpings ();

		    /* Reset the daily stats if it's a new day 
		       There has got to be a better way to do this.
		       Anyone know what it is?
		     */
		    tm = *gmtime (&curtime);
		    strftime (timebuf, sizeof (timebuf), "%H:%M", &tm);
		    if (!stricmp (timebuf, "21:46") && !donereset)
		    {
			reset_daily_stats ();
			donereset = 1;
		    }

		    /* So we dont keep doing it over and over for a minute */
		    if (!stricmp (timebuf, "21:47") && donereset)
			donereset = 0;
		}

		last_check = time (NULL);
	    }

	    if ((runflags & RUN_SAVEDBS) ||
		 curtime - last_update >= update_timeout)
	    {
		if (statserv_on == TRUE)
		    save_ss_dbase ();

		/* Check for any expired AKills */
		if (operserv_on == TRUE)
		{
		    expire_akills ();
		    save_os_dbase ();
		}

#ifdef DUMPCHANS
		dumpchans ();
#endif

		last_update = curtime;
		runflags &= ~RUN_SAVEDBS;
		snoop (s_OperServ, OS_DB_SYNC_COMPLETE);
	    }

	    i = (int)(long)sgets2 (inbuf, sizeof (inbuf), servsock);

	    if (i > 0)
	    {
		processing = 1;
		process ();
		processing = 0;
	    }

	    else if (i == 0)
	    {
		quitmsg = inbuf;
		log (ERR_ERROR, inbuf);
		runflags |= RUN_QUITTING;
	    }
	}

	if (!quitmsg)
	    quitmsg = "Terminating, reason unknown";

	log (ERR_QUIT, quitmsg);

	if (runflags & RUN_STARTED)
	    send_cmd (me.name, "SQUIT %s :%s", me.name, quitmsg);

        disconn (servsock);

	/* Should we reboot? Sleep, then restart. */
	if (wait_restart > 0 && !(runflags & RUN_SHUTDOWN))
	{
	    log (RPL_RESTARTING, duration (wait_restart));
	    close_log ();
	    if (debuglevel)
		close_debug();
	    sleep (wait_restart);
	    execve (OS_BINARY, av, envp);
	    return 1;
	}
	else
	{
	    close_log ();
	    if (debuglevel)
		close_debug();
	    remove (PIDFILE);
	    return 0;
	}
    }
    return 0;
}

/* Bring a new client onto the network. */
void intro_user (const char *nick, const char *user, const char *host,
		 const char *real, const char *modes)
{
#ifdef UNREAL3
    send_cmd (NULL, "NICK %s 1 %lu %s %s %s 1 %s * :%s", nick, time (NULL),
	user, host, me.name, modes, real);
#elif defined (BAHAMUT)
    send_cmd (NULL, "NICK %s 1 %lu %s %s %s %s 0 :%s", nick, time (NULL),
	modes, user, host, me.name, real);
#else
    send_cmd (NULL, "NICK %s 1 %lu %s %s %s 1 :%s", nick, time (NULL),
	user, host, me.name, real);
    send_cmd (nick, "MODE %s %s", nick, modes);
#endif
}

/* Check if we should ping a server */
void checkpings (void)
{
    int i;

    for (i = 0; i < servcnt; ++i)
    {
	if ((time (NULL) - servlist[i].lastping) > ping_delay)
	{
	    send_cmd (me.name, "PING %s %s", me.name, servlist[i].server);
	    servlist[i].lastping = time (NULL);
	}
    }
}
